from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPstartDate = ChildProtectionPlans.CPPstartDate
LAchildID = ChildProtectionPlans.LAchildID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate

CINdetails = CINTable.CINdetails
DateOfInitialCPC = CINdetails.DateOfInitialCPC

Section47 = CINTable.Section47


# define characteristics of rule
@rule_definition(
    code="2883",
    # module is table that seems central to the condition.
    module=CINTable.ChildProtectionPlans,
    message="There are more child protection plans starting than initial conferences taking place",
    affected_fields=[CPPstartDate, DateOfInitialCPC],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpp = data_container[ChildProtectionPlans]
    df_cin = data_container[CINdetails]
    df_47 = data_container[Section47]

    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # Within a Local Authority, count the number of <CPPStartDate> (N00105) where a date is present and within [Period_of_Census]. This value should be less than or equal to the sum of:
    # a) the count of <DateOfInitialCPC> (N00110) on CIN Details module where a date is present and within [Period_of_Census], plus
    # b) the count of <DateOfInitialCPC> on the S47 module where a date is present and within [Period_of_Census].

    # filter and count CPPstartDate
    present_cpp = df_cpp[CPPstartDate].notna()
    within_census_cpp = (df_cpp[CPPstartDate] >= collection_start) & (
        df_cpp[CPPstartDate] <= collection_end
    )
    df_cpp = df_cpp[present_cpp & within_census_cpp]
    num_cpp = len(df_cpp)

    # filter and count DateOfInitialCPC in CINdetails
    present_cin = df_cin[DateOfInitialCPC].notna()
    within_census_cin = (df_cin[DateOfInitialCPC] >= collection_start) & (
        df_cin[DateOfInitialCPC] <= collection_end
    )
    df_cin = df_cin[present_cin & within_census_cin]
    num_cin = len(df_cin)

    # filter and count DateOfInitialCPC in Section47
    present_47 = df_47[DateOfInitialCPC].notna()
    within_census_47 = (df_47[DateOfInitialCPC] >= collection_start) & (
        df_47[DateOfInitialCPC] <= collection_end
    )
    df_47 = df_47[present_47 & within_census_47]
    num_47 = len(df_47)

    if num_cpp > (num_cin + num_47):
        rule_context.push_la_level(
            rule_context.definition.code, rule_context.definition.message
        )
    else:
        pass


def test_validate():
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # collection_start is 01/04/2000
    )
    sample_cpp = pd.DataFrame(
        [  # num_cpp = 3
            {
                LAchildID: "child1",
                CPPstartDate: "26/05/2000",
            },
            {
                LAchildID: "child1",  # checks that values are considered independently and no grouping is done.
                CPPstartDate: "27/06/2000",
            },
            {
                LAchildID: "child2",
                CPPstartDate: "26/05/2004",  # not considered: out of period of census
            },
            {
                LAchildID: "child3",
                CPPstartDate: "26/05/2000",
            },
            {
                LAchildID: "child1",
                CPPstartDate: pd.NA,  # ignore: absent
            },
        ]
    )

    sample_cin = pd.DataFrame(
        [  # num_cin = 0
            {
                LAchildID: "child2",
                DateOfInitialCPC: "26/05/2004",  # not considered: out of period of census
            },
            {
                LAchildID: "child1",
                DateOfInitialCPC: pd.NA,  # ignore: absent
            },
        ]
    )

    sample_section47 = pd.DataFrame(
        [  # num_47 = 2
            {
                LAchildID: "child1",
                DateOfInitialCPC: "26/05/2000",
            },
            {
                LAchildID: "child1",  # checks that values are considered independently and no grouping is done.
                DateOfInitialCPC: "27/06/2000",
            },
        ]
    )

    sample_cpp[CPPstartDate] = pd.to_datetime(
        sample_cpp[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[DateOfInitialCPC] = pd.to_datetime(
        sample_cin[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47[DateOfInitialCPC] = pd.to_datetime(
        sample_section47[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            CINdetails: sample_cin,
            Section47: sample_section47,
            Header: sample_header,
        },
    )

    issues = result.la_issues
    assert issues == (
        "2883",
        "There are more child protection plans starting than initial conferences taking place",
    )



from .__api import CINTable, RuleDefinition, RuleType, YearConfig
from .__context import IssueLocator, RuleContext
from .__registry import rule_definition

__all__ = [
    "YearConfig",
    "RuleDefinition",
    "RuleType",
    "CINTable",
    "rule_definition",
    "RuleContext",
    "IssueLocator",
]



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.rules.cin2022_23.rule_8535Q import PersonDeathDate
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
PersonBirthDate = ChildIdentifiers.PersonBirthDate
ExpectedPersonBirthDate = ChildIdentifiers.ExpectedPersonBirthDate
LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="8525Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="Either Date of Birth or Expected Date of Birth must be provided (but not both)",
    affected_fields=[PersonBirthDate, ExpectedPersonBirthDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df = df.drop(columns=["ROW_ID"], errors="ignore")
    df.index.name = "ROW_ID"

    # Either Date of Birth or Expected Date of Birth must be provided (but not both)
    # condition_1 = (df[PersonBirthDate].isna() & df[ExpectedPersonBirthDate].isna())

    # condition_1 = (df[PersonBirthDate].isna()) & (df[ExpectedPersonBirthDate].isna())
    # condition_2 = df[PersonBirthDate].notna() & df[ExpectedPersonBirthDate].notna()
    mega_condition = (
        df[PersonBirthDate].isna() & df[ExpectedPersonBirthDate].notna()
    ) | (df[PersonBirthDate].notna() & df[ExpectedPersonBirthDate].isna())

    df_issues = df[~mega_condition].reset_index()

    # (LAchildID,PersonBirthDate,ExpectedPersonBirthDate) could have been used. However, in some failing conditions,
    # both (PersonBirthDate,ExpectedPersonBirthDate) can be null so their combination does not serve as a unique ID.
    # Since this is the ChildIdentifiers table and LAchildID is typically unique in it. We use that to serve as a last resort ID.

    link_id = tuple(
        zip(
            df_issues[LAchildID],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=ChildIdentifiers,
        columns=[PersonBirthDate, ExpectedPersonBirthDate],
        row_df=df_issues,
    )


def test_validate():
    fake_data_frame = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": "26/05/2000",
            },  # Fails because both DOB and expected DOB are present
            {
                "LAchildID": "child2",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": "26/05/2001",
            },  # Fails because both DOB and expected DOB are present
            {
                "LAchildID": "child4",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "26/05/1999",
            },
            {
                "LAchildID": "child4",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": pd.NA,
            },
            {
                "LAchildID": "child5",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": "25/05/2000",
            },  # Fails because both DOB and expected DOB are present
            {
                "LAchildID": "child6",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": pd.NA,
            },  # Fails because there is no DOB or expected DOB
        ]
    )

    result = run_rule(validate, {ChildIdentifiers: fake_data_frame})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [PersonBirthDate, ExpectedPersonBirthDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 4
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": ("child1",),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": ("child2",),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": ("child5",),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": ("child6",),
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8525Q"
    assert (
        result.definition.message
        == "Either Date of Birth or Expected Date of Birth must be provided (but not both)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPendDate = ChildProtectionPlans.CPPendDate
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8930'
    code="8930",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child Protection Plan End Date must fall within the census year",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[CPPendDate, ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildProtectionPlans]

    # ReferenceDate exists in the heder table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # If <CPPendDate> (N00115) is present, then<CPPendDate> (N00115) must fall within [Period_of_Census] inclusive
    failing_indices = df[
        (df[CPPendDate] < collection_start) | (df[CPPendDate] > reference_date)
    ].index

    rule_context.push_issue(
        table=ChildProtectionPlans, field=CPPendDate, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )

    fake_cpp = pd.DataFrame(
        [
            {
                CPPendDate: "01/03/2019"
            },  # 0 fail: March 1st is before April 1st, 2021. It is out of range
            {
                CPPendDate: "01/04/2021"
            },  # 1 pass: April 1st is within April 1st, 2021 to March 31st, 2022.
            {
                CPPendDate: "01/10/2022"
            },  # 2 fail: October 1st is after March 31st, 2022. It is out of range
            {
                CPPendDate: pd.NA
            },  # 2 fail: October 1st is after March 31st, 2022. It is out of range
        ]
    )

    # if date columns are involved, the validate function will be expecting them as dates so convert before passing them in.
    fake_cpp[CPPendDate] = pd.to_datetime(
        fake_cpp[CPPendDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    # Since the ReferenceDate comes from the Header column, we provide that also.
    result = run_rule(validate, {ChildProtectionPlans: fake_cpp, Header: fake_header})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.ChildProtectionPlans, CPPendDate, 0),
        IssueLocator(CINTable.ChildProtectionPlans, CPPendDate, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '8930' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8930"
    assert (
        result.definition.message
        == "Child Protection Plan End Date must fall within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

Disabilities = CINTable.Disabilities
Disability = Disabilities.Disability
LAchildID = Disabilities.LAchildID


# define characteristics of rule
@rule_definition(
    code="2888Q",
    rule_type=RuleType.QUERY,
    module=CINTable.Disabilities,
    message="Please check and either amend or provide a reason: Only one disability code is recorded per child and multiple disabilities should be recorded where possible.",
    affected_fields=[Disability],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Disabilities]

    # LOGIC
    # Excluding children where <Disability> (N00099) is ‘NONE’, within a local authority, one or more children should have more than one disability recorded.
    # Validation should be triggered at LA level, not child level, if all children who are recorded as having a disability have only 1 disability code recorded.

    # remove "NONE" values
    df = df[df[Disability] != "NONE"]

    disability_count = df.groupby(LAchildID)[Disability].count()
    # maximum number of disabilities recorded per child should be > 1

    if disability_count.max() <= 1:
        rule_context.push_la_level(
            rule_context.definition.code, rule_context.definition.message
        )
    else:
        pass


def test_validate():
    sample_disabilities = pd.DataFrame(
        [
            {
                LAchildID: "child0",
                Disability: "NONE",
            },  # child0 : disability_count would have been 2 and rule would have not been flagged if NONE was considered.
            {LAchildID: "child0", Disability: "NONE"},
            {LAchildID: "child1", Disability: "aaaa"},  # child1 : disability_count == 1
            {LAchildID: "child2", Disability: "bbbb"},  # child2 : disability_count == 1
            {LAchildID: "child2", Disability: pd.NA},
            {LAchildID: "child4", Disability: "cc"},  # child4 : disability_count == 1
        ]
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Disabilities: sample_disabilities})

    assert result.la_issues == (
        "2888Q",
        "Please check and either amend or provide a reason: Only one disability code is recorded per child and multiple disabilities should be recorded where possible.",
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

Assessments = CINTable.Assessments
AssessmentFactors = Assessments.AssessmentFactors


# define characteristics of rule
@rule_definition(
    code="8950Q",
    rule_type=RuleType.QUERY,
    module=CINTable.Assessments,
    message="Please check and either amend data or provide a reason: the assessment factors code '19A' should not be used ('19B' or '19C' should be used instead)",
    affected_fields=[AssessmentFactors],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Assessments].copy()

    with_af = df[df[AssessmentFactors].notna()]
    failing_indices = with_af[with_af[AssessmentFactors].str.upper() == "19A"].index

    rule_context.push_issue(
        table=Assessments, field=AssessmentFactors, row=failing_indices
    )


def test_validate():
    sample_assessments = pd.DataFrame(
        {"AssessmentFactors": [pd.NA, "19a", "19A", "18B", "19C"]}
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Assessments: sample_assessments})

    issues = list(result.issues)
    # Intended fail points in data.
    assert len(issues) == 2
    # Intended failures of test data by index.
    assert issues == [
        IssueLocator(CINTable.Assessments, AssessmentFactors, 1),
        IssueLocator(CINTable.Assessments, AssessmentFactors, 2),
    ]

    # Checks rule code and message are correct.
    assert result.definition.code == "8950Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: the assessment factors code '19A' should not be used ('19B' or '19C' should be used instead)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

Cindetails = CINTable.CINdetails
CINreferralDate = Cindetails.CINreferralDate
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="8600",
    module=CINTable.CINdetails,
    message="Child referral date missing or after data collection period",
    affected_fields=[CINreferralDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Cindetails]
    df_ref = data_container[Header]

    ref_data_series = df_ref[ReferenceDate]

    collection_start, collection_end = make_census_period(ref_data_series)

    # <CINreferralDate> (N00100) must be present and must be on or before <ReferenceDate> (N00603)
    condition = (df[CINreferralDate] > collection_end) | (df[CINreferralDate].isna())
    # Where a CINreferralDate exists check to see if it is after the end of the Census Period (collection_end)
    df = df[condition]

    failing_indices = df.index

    rule_context.push_issue(
        table=Cindetails, field=CINreferralDate, row=failing_indices
    )


def test_validate():
    sample_refdates = pd.to_datetime(
        [
            "2021-06-30",  # Pass - the date is before the collection_end of 31st March 2022
            "2022-01-25",  # Pass - the date is before the collection_end of 31st March 2022
            "2022-12-25",  # Fail - the date is after the collection_end of 31st March 2022
            pd.NA,  # Fail - the date field is blank and is therefore missing
            "2022-12-03",  # Fail - the date is after the collection_end of 31st March 2022
            "2021-08-20",  # Pass - the date is before the collection_end of 31st March 2022
            "2021-04-17",  # Pass - the date is before the collection_end of 31st March 2022
            "2001-01-25",  # Pass - the date is before the collection_end of 31st March 2022
            pd.NA,  # Fail - the date field is blank and is therefore missing
        ],
        format="%Y/%m/%d",
        errors="coerce",
    )

    fake_refdates = pd.DataFrame({CINreferralDate: sample_refdates})
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]
    )  # the census start date here will be 01/04/2021

    result = run_rule(validate, {Cindetails: fake_refdates, Header: fake_header})

    issues = list(result.issues)

    assert len(issues) == 4

    assert issues == [
        IssueLocator(CINTable.CINdetails, CINreferralDate, 2),
        IssueLocator(CINTable.CINdetails, CINreferralDate, 3),
        IssueLocator(CINTable.CINdetails, CINreferralDate, 4),
        IssueLocator(CINTable.CINdetails, CINreferralDate, 8),
    ]

    assert result.definition.code == "8600"
    assert (
        result.definition.message
        == "Child referral date missing or after data collection period"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID
UPNunknown = ChildIdentifiers.UPNunknown

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
ReferralNFA = CINdetails.ReferralNFA


@rule_definition(
    code="8772",
    module=CINTable.ChildIdentifiers,
    message="UPN unknown reason is UN7 (Referral with no further action) but at least one CIN details is a referral going on to further action",
    affected_fields=[
        UPNunknown,
        ReferralNFA,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_upn = data_container[ChildIdentifiers].copy()
    df_refs = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_upn.index.name = "ROW_ID"
    df_refs.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_upn.reset_index(inplace=True)
    df_refs.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <UPNunknown> (N00135) is UN7 then all of the CIN details must have <ReferralNFA> (N00112) = 1 or true

    df_merged = df_upn.merge(
        df_refs,
        on=["LAchildID"],
        how="left",
        suffixes=("_upn", "_refs"),
    )

    #  Get rows where UPNunknown is equal to 'UN7'
    condition_1 = df_merged[UPNunknown] == "UN7"

    #  Get rows where ReferralNFA is equal to either 'True' or '1'
    trueor1 = ["true", "1"]
    condition_2 = df_merged[ReferralNFA].str.lower().isin(trueor1)

    # Combine the results of the two rows
    df_merged = df_merged[condition_1 & ~condition_2].reset_index()

    # create an identifier for each error instance.
    df_merged["ERROR_ID"] = tuple(zip(df_merged[LAchildID], df_merged[ReferralNFA]))

    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_upn_issues = (
        df_upn.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_upn")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_refs_issues = (
        df_refs.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_refs")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[UPNunknown], row_df=df_upn_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReferralNFA], row_df=df_refs_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_upn = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "UPNunknown": "UN6",  # Ignore: code is not UN7
            },
            {
                "LAchildID": "child2",
                "UPNunknown": "UN7",
            },
            {
                "LAchildID": "child3",
                "UPNunknown": "UN7",
            },
            {
                "LAchildID": "child4",
                "UPNunknown": "UN7",
            },
            {
                "LAchildID": "child5",
                "UPNunknown": "UN4",  # Ignore: code is not UN7
            },
        ]
    )

    sample_refs = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "ReferralNFA": "True",  # 0 Ignore: upn is not UN7
            },
            {
                "LAchildID": "child2",
                "ReferralNFA": "1",  # 1 Pass as Referral Code is True
            },
            {
                "LAchildID": "child3",
                "ReferralNFA": pd.NA,  # 2 Fail as Referral Code is NULL
            },
            {
                "LAchildID": "child4",
                "ReferralNFA": "nottrue",  # 3 Fail: Referral Code is neither "1" not "true"
            },
            {
                "LAchildID": "child5",
                "ReferralNFA": "True",  # 4 Ignore: upn is not UN7
            },
        ]
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_upn,
            CINdetails: sample_refs,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2

    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values.
    issues = issues_list[1]

    # get table name and check it. Replace CINdetails with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace ReferralNFA with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [ReferralNFA]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df

    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)

    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    pd.NA,  # ReferralNFA
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",  # ChildID
                    "nottrue",  # ReferralNFA
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8772' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8772"
    assert (
        result.definition.message
        == "UPN unknown reason is UN7 (Referral with no further action) but at least one CIN details is a referral going on to further action"
    )



import pandas as pd

from .utils import get_values


# initialize all data sets as empty dataframes with columns names
# whenever a child is created, it should add a row to each table where it exists.
# tables should be attributes of a class that are accessible to the methods in create_child.
class XMLtoCSV:
    """
    A class to convert data input as XML into CSV/DataFrame format for validation. Uses
    ElementTree to parse the XML for each child, adding their data to relevant fields and tables.

    :param DataFrame Header: DataFrame of fields for the Header table for validation,
        to be populated with children's data from XML input.
    :param DataFrame ChildIdentifiers: DataFrame of fields for the ChildIdentifiers table for validation.
        to be populated with children's data from XML input.
    :param DataFrame ChildCharacteristics: DataFrame of fields for the ChildCharacterisitcs table for validation.
        to be populated with children's data from XML input.
    :param DataFrame Disabilities: DataFrame of fields for the Disabilities table for validation.
        to be populated with children's data from XML input.
    :param DataFrame CINdetails: DataFrame of fields for the CINdetails table for validation.
        to be populated with children's data from XML input.
    :param DataFrame Assessments: DataFrame of fields for the Assessments table for validation.
        to be populated with children's data from XML input.
    :param DataFrame CINplanDates: DataFrame of fields for the CINplanDates table for validation.
        to be populated with children's data from XML input.
    :param DataFrame Section47: DataFrame of fields for the Section47 table for validation.
        to be populated with children's data from XML input.
    :param DataFrame ChildProtectionPlans: DataFrame of fields for the ChildProtectionPlans table for validation.
        to be populated with children's data from XML input.
    :param DataFrame Reviews: DataFrame of fields for the Reviews table for validation.
        to be populated with children's data from XML input.
    :param list id_cols: List of columns containing IDs that can be used to merge tables.
    """

    # define column names from CINTable object.
    Header = pd.DataFrame(
        columns=[
            "Collection",
            "Year",
            "ReferenceDate",
            "SourceLevel",
            "LEA",
            "SoftwareCode",
            "Release",
            "SerialNo",
            "DateTime",
        ]
    )
    ChildIdentifiers = pd.DataFrame(
        columns=[
            "LAchildID",
            "UPN",
            "FormerUPN",
            "UPNunknown",
            "PersonBirthDate",
            "ExpectedPersonBirthDate",
            "GenderCurrent",
            "Sex",
            "PersonDeathDate",
        ]
    )
    ChildCharacteristics = pd.DataFrame(
        columns=[
            "LAchildID",
            "Ethnicity",
        ]
    )
    Disabilities = pd.DataFrame(columns=["LAchildID", "Disability"])

    CINdetails = pd.DataFrame(
        columns=[
            "LAchildID",
            "CINdetailsID",
            "CINreferralDate",
            "ReferralSource",
            "PrimaryNeedCode",
            "CINclosureDate",
            "ReasonForClosure",
            "DateOfInitialCPC",
            "ReferralNFA",
        ]
    )
    Assessments = pd.DataFrame(
        columns=[
            "LAchildID",
            "CINdetailsID",
            "AssessmentID",
            "AssessmentActualStartDate",
            "AssessmentInternalReviewDate",
            "AssessmentAuthorisationDate",
            "AssessmentFactors",
        ]
    )
    AssessmentFactorsList = pd.DataFrame(
        columns=[
            "LAchildID",
            "CINdetailsID",
            "AssessmentID",
            "AssessmentFactor",
        ]
    )
    CINplanDates = pd.DataFrame(
        columns=["LAchildID", "CINdetailsID", "CINPlanStartDate", "CINPlanEndDate"]
    )
    Section47 = pd.DataFrame(
        columns=[
            "LAchildID",
            "CINdetailsID",
            "S47ActualStartDate",
            "InitialCPCtarget",
            "DateOfInitialCPC",
            "ICPCnotRequired",
        ]
    )
    ChildProtectionPlans = pd.DataFrame(
        columns=[
            "LAchildID",
            "CINdetailsID",
            "CPPID",
            "CPPstartDate",
            "CPPendDate",
            "InitialCategoryOfAbuse",
            "LatestCategoryOfAbuse",
            "NumberOfPreviousCPP",
        ]
    )
    Reviews = pd.DataFrame(
        columns=["LAchildID", "CINdetailsID", "CPPID", "CPPreviewDate"]
    )

    id_cols = ["LAchildID", "CINdetailsID", "AssessmentID", "CPPID"]

    def __init__(self, root):
        """
        Initialises XMLtoCSV class, creates header, and iterates through input XML for every Child field
        in the Children field.

        :param xml root: root of the CIN XML data
        :returns: Generates 10 dataframes containing the child info from the CIN XML fed into it.
        """

        header = root.find("Header")
        self.Header = self.create_Header(header)

        children = root.find("Children")
        for child in children.findall("Child"):
            self.create_child(child)

    # for each table, column names should attempt to find their value in the child.
    # if not found, they should assign themselves to NaN

    def create_child(self, child):
        """
        For every child, use class methods to extract relevant fields from
        input XML and append its data to appropriate DataFrames.
        """

        # at the start of every child, reset the value of LAchildID
        self.LAchildID = None

        self.create_ChildIdentifiers(child)
        # LAchildID has been created. It can be used in the functions below.
        self.create_ChildCharacteristics(child)

        # CINdetailsID needed
        self.create_CINdetails(child)

        # CINdetails and CPPID needed
        self.create_ChildProtectionPlans(child)
        self.create_Reviews(child)

    # TODO get column names from the CINTable object instead of writing them out as strings?
    def create_Header(self, header):
        """Extracts header data from XML, run once as only one row is needed for the header.
        Exists once per census return.

        :param object header: The element with the "Header" tag in the input XML
        :returns: A DataFrame of the Header table extracted from input XML.
        :rtype: DataFrame
        """

        header_dict = {}

        collection_details = header.find("CollectionDetails")
        collection_elements = ["Collection", "Year", "ReferenceDate"]
        header_dict = get_values(collection_elements, header_dict, collection_details)

        source = header.find("Source")
        source_elements = [
            "SourceLevel",
            "LEA",
            "SoftwareCode",
            "Release",
            "SerialNo",
            "DateTime",
        ]
        header_dict = get_values(source_elements, header_dict, source)

        header_df = pd.DataFrame.from_dict([header_dict])
        return header_df

    def create_ChildIdentifiers(self, child):
        """
        Populates the ChildIdentifiers table. One ChildIdentifiers block exists per child in CIN XML

        :param xml child: 'child' element from the XML input. Each contains the full information per child.
        :returns: DataFrame of data for an individual child for the ChildIdentifiers Table.
        :rtype: DataFrame
        """

        # pick out the values of relevant columns
        # add to the global attribute
        identifiers_dict = {}

        identifiers = child.find("ChildIdentifiers")
        elements = self.ChildIdentifiers.columns
        identifiers_dict = get_values(elements, identifiers_dict, identifiers)

        self.LAchildID = identifiers_dict.get("LAchildID", pd.NA)

        identifiers_df = pd.DataFrame.from_dict([identifiers_dict])
        self.ChildIdentifiers = pd.concat(
            [self.ChildIdentifiers, identifiers_df], ignore_index=True
        )

    def create_ChildCharacteristics(self, child):
        """Populates the ChildCharacteristics table. One ChildCharacteristics block exists per child in CIN XML

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the ChildCharacteristics Table.
        :rtype: DataFrame
        """
        # assign LAChild ID
        characteristics_dict = {"LAchildID": self.LAchildID}

        characteristics = child.find("ChildCharacteristics")
        columns = self.ChildCharacteristics.columns
        # select only columns whose values typically exist in this xml block.
        # remove id_cols which tend to come from other blocks or get generated at runtime.
        elements = list(set(columns).difference(set(self.id_cols)))

        characteristics_dict = get_values(
            elements, characteristics_dict, characteristics
        )

        characteristics_df = pd.DataFrame.from_dict([characteristics_dict])
        self.ChildCharacteristics = pd.concat(
            [self.ChildCharacteristics, characteristics_df], ignore_index=True
        )

        # The disabilities block for a child is found within a ChildCharacteristics block.
        self.create_Disabilities(characteristics)

    def create_Disabilities(self, characteristics):
        """
        Populates Disabilites table
        """
        disabilities_list = []
        columns = self.Disabilities.columns
        elements = list(set(columns).difference(set(self.id_cols)))
        # get the Disabilities block
        disabilities = characteristics.find("Disabilities")
        if disabilities is not None:
            # Only run this if a "Disabilities" xml block has been found
            for disability in disabilities:
                disability_dict = {
                    "LAchildID": self.LAchildID,
                }
                disability_dict = get_values(elements, disability_dict, disability)
                disability_dict["Disability"] = disability.text
                disabilities_list.append(disability_dict)

            disabilities_df = pd.DataFrame(disabilities_list)
            self.Disabilities = pd.concat(
                [self.Disabilities, disabilities_df], ignore_index=True
            )

    # CINdetailsID needed
    def create_CINdetails(self, child):
        """Populates the CINdetails table. Multiple CIN details blocks can exist in one child.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the CINdetails Table.
        :rtype: DataFrame
        """

        cin_details_list = []
        columns = self.CINdetails.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        # TODO should we imitate DfE generator where the ID count for the first child is 1?
        self.CINdetailsID = 0

        cin_details = child.findall("CINdetails")
        for cin_detail in cin_details:
            self.CINdetailsID += 1
            cin_detail_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
            }

            cin_detail_dict = get_values(elements, cin_detail_dict, cin_detail)
            cin_details_list.append(cin_detail_dict)

            # functions that should use the CINdetailsID before it is incremented.
            self.create_Assessments(cin_detail)
            self.create_CINplanDates(cin_detail)
            self.create_Section47(cin_detail)
            self.create_ChildProtectionPlans(cin_detail)

        cin_details_df = pd.DataFrame(cin_details_list)
        self.CINdetails = pd.concat(
            [self.CINdetails, cin_details_df], ignore_index=True
        )

    def create_Assessments(self, cin_detail):
        """Populates the assessments table. Multiple Assessments blocks can exist in one CINdetails block.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the Assessments Table.
        :rtype: DataFrame
        """

        assessments_list = []
        columns = self.Assessments.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        self.AssessmentID = 0
        assessments = cin_detail.findall("Assessments")

        for assessment in assessments:
            self.AssessmentID += 1
            assessment_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
                "AssessmentID": self.AssessmentID,
            }

            assessment_dict = get_values(elements, assessment_dict, assessment)

            # the get_values function will not find AssessmentFactors on that level so we retrieve these separately.
            assessment_factors = assessment.find("FactorsIdentifiedAtAssessment")
            assessment_factors_list = []
            assessment_columns = self.AssessmentFactorsList.columns
            assessment_elements = list(
                set(assessment_columns).difference(set(self.id_cols))
            )

            if assessment_factors is not None:
                # if statement handles the non-iterable NoneType that .find produces if the element is not present.
                if len(assessment_factors) != 0:
                    # checks that the assessment facros block has any assessment factors in it as
                    # some close without any in.
                    for factor in assessment_factors:
                        assessment_factors_dict = {
                            "LAchildID": self.LAchildID,
                            "CINdetailsID": self.CINdetailsID,
                            "AssessmentID": self.AssessmentID,
                        }
                        assessment_factors_dict = get_values(
                            assessment_elements, assessment_factors_dict, factor
                        )
                        assessment_factors_dict["AssessmentFactor"] = factor.text
                        assessment_factors_list.append(assessment_factors_dict)
                    assessment_factors_df = pd.DataFrame(assessment_factors_list)
                    self.AssessmentFactorsList = pd.concat(
                        [self.AssessmentFactorsList, assessment_factors_df],
                        ignore_index=True,
                    )
                    assessment_dict["AssessmentFactors"] = assessment_factors_df[
                        "AssessmentFactor"
                    ].tolist()

            assessments_list.append(assessment_dict)

        assessments_df = pd.DataFrame(assessments_list)
        print(assessments_df)
        self.Assessments = pd.concat(
            [self.Assessments, assessments_df], ignore_index=True
        )

    def create_CINplanDates(self, cin_detail):
        """
        Populates the CINplanDates table. Multiple CINplanDates blocks can exist in one CINdetails block.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the CINplanDates Table.
        :rtype: DataFrame
        """

        dates_list = []
        columns = self.CINplanDates.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        dates = cin_detail.findall("CINPlanDates")
        for date in dates:
            date_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
            }
            date_dict = get_values(elements, date_dict, date)
            dates_list.append(date_dict)

        dates_df = pd.DataFrame(dates_list)
        self.CINplanDates = pd.concat([self.CINplanDates, dates_df], ignore_index=True)

    def create_Section47(self, cin_detail):
        """
        Populates the Section47 table. Multiple Section47 blocks can exist in one CINdetails block.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the Section47 Table.
        :rtype: DataFrame
        """

        sections_list = []
        columns = self.Section47.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        sections = cin_detail.findall("Section47")
        for section in sections:
            section_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
            }
            section_dict = get_values(elements, section_dict, section)
            sections_list.append(section_dict)

        sections_df = pd.DataFrame(sections_list)
        self.Section47 = pd.concat([self.Section47, sections_df], ignore_index=True)

    # CINdetails and CPPID needed
    def create_ChildProtectionPlans(self, cin_detail):
        """
        Populates the ChildProtectionPlans table. Multiple ChildProtectionPlans blocks can exist in one CINdetails block.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the ChildProtectionPlans Table.
        :rtype: DataFrame
        """

        plans_list = []
        columns = self.ChildProtectionPlans.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        # imitate DfE generator where the first counted thing starts from 1.
        self.CPPID = 0

        plans = cin_detail.findall("ChildProtectionPlans")
        for plan in plans:
            self.CPPID += 1
            plan_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
                "CPPID": self.CPPID,
            }
            plan_dict = get_values(elements, plan_dict, plan)
            plans_list.append(plan_dict)

            # functions that should use CPPID before it is incremented
            self.create_Reviews(plan)

        plans_df = pd.DataFrame(plans_list)
        self.ChildProtectionPlans = pd.concat(
            [self.ChildProtectionPlans, plans_df], ignore_index=True
        )

    def create_Reviews(self, plan):
        """
        Populates the ChildIdentifiers table. Multiple Reviews blocks can exist in one Reviews block.

        :param xml child: 'child' element from the XML input
        :returns: DataFrame of data for an individual child for the Reviews Table.
        :rtype: DataFrame
        """

        reviews_list = []
        columns = self.Reviews.columns
        elements = list(set(columns).difference(set(self.id_cols)))

        reviews = plan.findall("Reviews[CPPreviewDate]")
        for review in reviews:
            review_dict = {
                "LAchildID": self.LAchildID,
                "CINdetailsID": self.CINdetailsID,
                "CPPID": self.CPPID,
            }
            review_dict = get_values(elements, review_dict, review)

            reviews_list.append(review_dict)

        reviews_df = pd.DataFrame(reviews_list)
        self.Reviews = pd.concat([self.Reviews, reviews_df], ignore_index=True)


"""
Sidenote: Fields absent from the fake_CIN_data.xml
- Assessments
- CINPlanDates
- Section47
- ChildProtectionPlans
- Reviews
"""



import importlib
from typing import Iterable

from cin_validator.rule_engine import RuleDefinition, YearConfig


def check_duplicate_rules(new_funcs: dict, funcs_so_far: dict) -> None:
    duplicate_funcs = set(new_funcs.keys()) & set(funcs_so_far.keys())
    if duplicate_funcs:
        raise ValueError(f"Rule with code {duplicate_funcs} already exists")


def extract_validator_functions(
    file_paths: Iterable, marker: str = "__rule_def__"
) -> dict[str, RuleDefinition]:
    """
    :param list file_paths: list of file paths to extract functions from.
    :param str marker: marker to identify functions to extract.

    :return: functions extracted from files.
    :rtype: dict
    """
    validator_funcs: dict[str, RuleDefinition] = {}
    for path in file_paths:
        if path.stem == "__init__":
            continue
        try:
            rule_content = importlib.import_module(
                # for example, path.parent.stem == "cin2022_23" and path.stem == "rule_100.py"
                f"cin_validator.rules.{path.parent.stem}.{path.stem}"
            )
        except ModuleNotFoundError:
            # in the case where the file itself is passed in, rather than the directory
            rule_content = importlib.import_module(f"{path.stem}")

        validator_func = {
            str(element.__rule_def__.code): element.__rule_def__
            for _, element in vars(rule_content).items()
            if hasattr(element, "__rule_def__")
        }

        check_duplicate_rules(validator_func, validator_funcs)

        validator_funcs.update(validator_func)
    return validator_funcs


def update_validator_functions(
    prev_validator_funcs, this_year_config: YearConfig
) -> dict:
    """
    :param dict prev_registry: previous year's registry.
    :param dict this_year_config: codes of rules that have been added or deleted.

    :return: valid validator functions according to config.
    :rtype: dict
    """
    # Rules present in both will be updated to this year's version. rules present in only this year will be added.
    updated_validator_funcs = prev_validator_funcs | this_year_config.added_or_modified
    # delete rules by their rules codes, if specified.
    for deleted_rule in this_year_config.deleted:
        del updated_validator_funcs[deleted_rule]
    return updated_validator_funcs


def get_year_ruleset(collection_year: str) -> dict[str, RuleDefinition]:
    """
    Gets the registry of validation rules for the year specified in the metadata.
    """
    # for example, convert "2023" to "lac2022_23"
    ruleset = f"cin{int(collection_year)-1}_{collection_year[2:4]}"

    module = importlib.import_module(f"cin_validator.rules.{ruleset}")
    registry = getattr(module, "registry")

    return registry



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
Section47 = CINTable.Section47
Assessments = CINTable.Assessments

AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
S47ActualStartDate = Section47.S47ActualStartDate
ReferralNFA = CINdetails.ReferralNFA
DateOfInitialCPC = CINdetails.DateOfInitialCPC

LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8831'
    code="8831",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Activity is recorded against a case marked as a referral with no further action",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        AssessmentActualStartDate,
        AssessmentAuthorisationDate,
        S47ActualStartDate,
        DateOfInitialCPC,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_ass = data_container[Assessments].copy()
    df_47 = data_container[Section47].copy()
    df_cin = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"
    df_ass.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)
    df_ass.reset_index(inplace=True)

    # If a <CINdetails> module has <ReferralNFA> (N00112) = true or 1, then it cannot have any of the following:
    # <AssessmentActualStartDate> (N00159)
    # <AssessmentAuthorisationDate> (N00160)
    # <S47ActualStartDate> (N00148)
    # <DateOfInitialCPC> (N00110)

    df_cin = df_cin[df_cin[ReferralNFA].isin(["true", "1"])]

    # Check columns in Section47 table
    df_cin_47 = df_cin.merge(
        df_47,
        on=[
            "LAchildID",
            "CINdetailsID",
        ],
        how="left",
        suffixes=["_cin", "_47"],
    )

    # filter out rows that have an S47ActualStartDate or DateOfInitialCPC from the CINdetails module
    condition_1 = (
        df_cin_47["DateOfInitialCPC_cin"].notna()
        | df_cin_47[S47ActualStartDate].notna()
    )
    df_cin_47 = df_cin_47[condition_1]

    df_cin_47["ERROR_ID"] = tuple(zip(df_cin_47[LAchildID], df_cin_47[CINdetailsID]))

    df_47_issues = (
        df_47.merge(df_cin_47, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    cin_issues_47 = (
        df_cin.merge(df_cin_47, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Check columns in Assessments table
    df_cin_ass = df_cin.merge(
        df_ass,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_ass"],
    )
    # filter out rows that have an AssessmentActualStartDate or AssessmentAuthorisationDate
    condition_2 = df_cin_ass[AssessmentActualStartDate].notna()
    condition_3 = df_cin_ass[AssessmentAuthorisationDate].notna()
    df_cin_ass = df_cin_ass[condition_2 | condition_3]
    df_cin_ass["ERROR_ID"] = tuple(zip(df_cin_ass[LAchildID], df_cin_ass[CINdetailsID]))
    df_ass_issues = (
        df_ass.merge(df_cin_ass, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    cin_issues_ass = (
        df_cin.merge(df_cin_ass, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cin_issues = pd.concat([cin_issues_47, cin_issues_ass])
    # in case a value was flagged in both table combinations, it'll exist twice so deduplicate df_cin_issues
    df_cin_issues.drop_duplicates("ERROR_ID")
    df_cin_issues.reset_index(inplace=True, drop=True)

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=CINdetails, columns=[DateOfInitialCPC], row_df=df_cin_issues
    )
    rule_context.push_type_2(
        table=Assessments, columns=[LAchildID], row_df=df_ass_issues
    )
    rule_context.push_type_2(table=Section47, columns=[LAchildID], row_df=df_47_issues)


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_section47 = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child2",  # fails for having a module
                "CINdetailsID": "cinID2",
                "DateOfInitialCPC": pd.NA,
                "S47ActualStartDate": "01/01/2000",
            },
            {
                "LAchildID": "child2",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID4",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 fail: has AssessmentAuthorisationDate
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID1",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child2",  # 1 fail: has S47ActualStartDate
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID2",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child3",  # 2 fail: has AssessmentActualStartDate
                "DateOfInitialCPC": pd.NA,
                "CINdetailsID": "cinID3",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child4",
                "DateOfInitialCPC": "28/05/2000",  # 3 fails for having initial cpc
                "CINdetailsID": "cinID4",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2000",
                "CINdetailsID": "cinID2",
                "ReferralNFA": "false",  # 4 ignore
            },
            {
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2003",
                "CINdetailsID": "cinID8",
                "ReferralNFA": "false",  # 5 ignore
            },
            {  # 6 pass
                "LAchildID": "child3",
                "DateOfInitialCPC": "14/03/2001",
                "CINdetailsID": "cinID4",
                "ReferralNFA": "false",  # 6 ignore
            },
        ]
    )
    sample_ass = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # fails for having a module
                "CINdetailsID": "cinID1",
                "AssessmentAuthorisationDate": "01/01/2000",
                "AssessmentActualStartDate": pd.NA,
            },
            {
                "LAchildID": "child3",  # fails for having a module
                "CINdetailsID": "cinID3",
                "AssessmentAuthorisationDate": "01/01/2000",
                "AssessmentActualStartDate": "01/01/2000",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID2",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": pd.NA,
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID3",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": pd.NA,
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID4",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": pd.NA,
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.

    sample_cin_details["DateOfInitialCPC"] = pd.to_datetime(
        sample_cin_details["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47["DateOfInitialCPC"] = pd.to_datetime(
        sample_section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47["S47ActualStartDate"] = pd.to_datetime(
        sample_section47["S47ActualStartDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_ass["AssessmentAuthorisationDate"] = pd.to_datetime(
        sample_ass["AssessmentAuthorisationDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_ass["AssessmentActualStartDate"] = pd.to_datetime(
        sample_ass["AssessmentActualStartDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            Section47: sample_section47,
            CINdetails: sample_cin_details,
            Assessments: sample_ass,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 3
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 0 will contain the CINdetails columns because that's the first thing pushed above.
    issues = issues_list[0]

    # get table name and check it. Replace Section47 with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace DateOfInitialCPC  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 4 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 4
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    "cinID2",  # CINdetailsID
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child4",  # ChildID
                    "cinID4",  # CINdetailsID
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID3",  # CINdetailsID
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8831' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8831"
    assert (
        result.definition.message
        == "Activity is recorded against a case marked as a referral with no further action"
    )



from dataclasses import dataclass
from typing import List

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleDefinition
from cin_validator.utils import create_issue_locs


@dataclass(frozen=True, eq=True)
class IssueLocator:
    """
    Dataclass used to specify definite locations of issues that should be highlighted in the user's data.

    :param CINTable-object table: Contains the name of the module/table in erorr for
        a validation rule.
    :param str field: The name of the field/column of an error in validation.
    :param int row: The index/row number of an error.
    :returns: IssueLocator object containing table, field, and row of validation errors.
    :rtype: IssueLocator object.
    """

    table: CINTable
    field: str
    row: int


@dataclass(frozen=True, eq=True)
class Type1:
    """Dataclass to define issue locations when more than one column is involved."""

    table: CINTable
    columns: List[str]
    row_df: pd.DataFrame


class RuleContext:
    """
    The RuleContext class includes methods that define how error locations
    should be stored per validation rule.

    >Type 0 rules contain 1 table and 1 column.
    >Type 1 rules contain multiple columns, but one table and no merges.
    >Type 2 rules contain multiple tables and columns.
    >Type 3 rules contain  one table but use merges to check values by group,
        e.g. a CINplan group.
    >LA level rules contain checks for a whole local authority.
    """

    def __init__(self, definition: RuleDefinition):
        """
        Initialises RuleContext class.

        :param RuleDefinition-object definition: Member of the rule definition dataclass,
            contains information about each validation rule.
        :param list issues: Empty list to be populated with type 0 and 1 issues.
        :param list type2_issues: Empty list to be populated with type 2 issues.
        :param list type3_issues: Empty list to be populated with type 3 issues.
        """

        self.__definition = definition

        self.__issues: list = []
        self.__type1_issues: list = []
        self.__type2_issues: list = []
        self.__type3_issues: list = []
        self.__la_issues: list = []

    @property
    def definition(self):
        """
        Used to call information about validation rules.

        :returns: Object containing information about each validation rule.
        :rtype: RuleDefinition object.
        """

        return self.__definition

    # TODO create list of rules according to types to prevent checking all attributes each time a rule is run.
    # Possibly classify rule code by adding it to a list of rules with a similar type, when push is done.

    # METHODS THAT DEFINE HOW ERROR LOCATIONS SHOULD BE STORED PER RULE STRUCTURE
    def push_issue(self, table, field, row):
        """
        For rules that check only a single column.

        :param CINTable-object table: the table a validation error ocurred in.
        :param CINTable-object column: the column a validation error ocurred in.
        :param DataFrame row_df: errors for a validation rule by table.
        :returns: information to locate validation errors in original data.
        :rtype: list of IssueLocator objects.
        """

        for i in row:
            self.__issues.append(IssueLocator(table, field, i))

    def push_type_1(self, table, columns, row_df):
        """
        For rules that check multiple columns in a single table, no merge involved.

        :param CINTable-object table: the table a validation error ocurred in.
        :param CINTable-object column: the column a validation error ocurred in.
        :param DataFrame row_df: the errors for a validation rule by table.
        :returns: information to locate validation errors in original data.
        :rtype: dataclass object
        """

        self.__type1_issues = Type1(table, columns, row_df)

    def push_type_2(self, table, columns, row_df):
        """
        For rules that check multiple columns across multiple tables.

        :param CINTable-object table: the table a validation error ocurred in.
        :param CINTable-object column:the column a validation error ocurred in.
        :param DataFrame row_df: the errors for a validation rule by table.
        :returns: information to locate validation errors in original data.
        :rtype: list of dataclass objects
        """

        table_tuple = Type1(table, columns, row_df)
        self.__type2_issues.append(table_tuple)

    def push_type_3(self, table, columns, row_df):
        """
        For rules that check values in a group with respect to each other.

        :param CINTable-object table: the table a validation error ocurred in.
        :param CINTable-object column: the column a validation error ocurred in.
        :param DataFrame row_df: the errors for a validation rule by table.
        :returns: information to locate validation errors in original data.
        :rtype: list of dataclass objects
        """

        table_tuple = Type1(table, columns, row_df)
        self.__type3_issues.append(table_tuple)

    def push_la_level(self, rule_code, rule_description):
        """
        For rules that check relationships across the whole local authority

        :param CINTable-object table: the table a validation error ocurred in.
        :param CINTable-object column: the column a validation error ocurred in.
        :param DataFrame row_df: the errors for a validation rule by table.
        :returns: information to locate validation errors in original data.
        :rtype: list of tuples
        """

        self.__la_issues = (rule_code, rule_description)

    # PROPERTIES FOR TEST_VALIDATE FUNCTIONS
    @property
    def issues(self):
        return self.__issues

    @property
    def type1_issues(self):
        return self.__type1_issues

    @property
    def type2_issues(self):
        return self.__type2_issues

    @property
    def type3_issues(self):
        return self.__type3_issues

    @property
    def la_issues(self):
        return self.__la_issues

    # PROPERTIES FOR CREATING THE ERROR REPORT
    @property
    def type_zero_issues(self):
        """
        Expands issues object into a dataframe where each row represents a location in the data
        by a unique table-column-index combination.

        :returns: DataFrame that contains failing locations of rules that involve only 1 column
        :rtype: DatFrame
        """

        if len(self.__issues) != 0:
            locator_dicts = []
            for locator in self.__issues:
                # convert every IssueLocator object to a dictionary
                locator_as_dict = {
                    "tables_affected": str(locator.table)[9:],
                    "columns_affected": str(locator.field),
                    "ROW_ID": str(locator.row),
                }
                locator_dicts.append(locator_as_dict)
            # create a df that contains data from all issue_locators generated by the rule.
            df_issue_locs = pd.DataFrame(locator_dicts)
            return df_issue_locs

        else:
            # for non-type0 rules, do this
            return []

    @property
    def type_one_issues(self):
        """
        Expands type1 issue object into a dataframe where each row represents a location in the data
        by a unique table-column-index combination.

        :returns: DataFrame that contains failing locations of rules that involve only 1 table
            and multiple columns.
        :rtype: DatFrame
        """
        try:
            # if it is a type1 rule i.e __type1_issues.row_df exists, do this.
            issues = self.__type1_issues
            df_issue_locs = create_issue_locs(issues)
            return df_issue_locs

        except:
            # all non-type1 rules run this.
            return []

    # type_one_issues and type_two_issues, though similar, should be left apart for readability.
    @property
    def type_two_issues(self):
        """
        Expands type2 issue object into a dataframe where each row represents a location in the data
        by a unique table-column-index combination

        :returns: DataFrame that contains failing locations of rules that involve multiple tables.
        :rtype: DataFrame
        """

        try:
            # if it is a type2 rule i.e __type2_issues.row_df exists, do this.
            issues_per_table = self.__type2_issues
            df_issue_locs_lst = []
            for issues in issues_per_table:
                # create a dataframe of issue locations for each table.
                df_issue_loc_table = create_issue_locs(issues)
                # append all table dataframes to a list.
                df_issue_locs_lst.append(df_issue_loc_table)
            # generate a dataframe that contains that data of all tables involved.
            df_issue_locs = pd.concat(df_issue_locs_lst, ignore_index=True)
            return df_issue_locs

        except:
            # all non-type2 rules run this.
            return []

    @property
    def type_three_issues(self):
        """
        Expands type3 issue object into a dataframe where each row represents a location in the data
        by a unique table-column-index combination.

        :returns: DataFrame that contains failing locations of rules that involve errors within groups.
        :rtype: DataFrame
        """

        try:
            # if it is a type3 rule i.e __type3_issues.row_df exists, do this.
            issues_per_table = self.__type3_issues
            df_issue_locs_lst = []
            for issues in issues_per_table:
                # create a dataframe of issue locations for each table.
                df_issue_loc_table = create_issue_locs(issues)
                # append all table dataframes to a list.
                df_issue_locs_lst.append(df_issue_loc_table)
            # generate a dataframe that contains that data of all tables involved.
            df_issue_locs = pd.concat(df_issue_locs_lst, ignore_index=True)
            return df_issue_locs

        except:
            # all non-type3 rules return this.
            return []

    @property
    def la_level_issues(self):
        """
        Creates DataFrame of return level validation errors.

        :returns: DataFrame containing rule code/description of la-level rules that the data failed.
        :rtype: DataFrame
        """
        try:
            code, desc = self.__la_issues
            la_df = pd.DataFrame(
                [{"rule_code": code, "rule_description": desc, "la_level": True}]
            )
            return la_df
        except:
            return []



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
CINclosureDate = CINdetails.CINclosureDate
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
LAchildID = Assessments.LAchildID
CINdetailsAssID = Assessments.CINdetailsID


# define characteristics of rule
@rule_definition(
    code="8867",
    module=CINTable.CINdetails,
    message="CIN episode is shown as closed, however Assessment is not shown as completed",
    affected_fields=[
        CINclosureDate,
        AssessmentAuthorisationDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_cind = data_container[CINdetails].copy()
    df_ass = data_container[Assessments].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cind.index.name = "ROW_ID"
    df_ass.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cind.reset_index(inplace=True)
    df_ass.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <CINclosureDate> (N00102) is present then all instances of the <Assessments> group must include the <AssesssmentAuthorisationDate> (N00160)
    # Issues dfs should return rows where an AssessmentAuthorisationDate exists if a CINclosureDate has been recorded for that child.

    #  Create dataframes which only have rows with a CINclosureDate and an AssessmentAuthorisation Date and which should have one plan per row.
    df_cind = df_cind[df_cind[CINclosureDate].notna()]

    #  Merge tables to get corresponding CP plan group and reviews
    df_merged = df_cind.merge(
        df_ass,
        left_on=["LAchildID", "CINdetailsID"],
        right_on=["LAchildID", "CINdetailsID"],
        how="inner",
        suffixes=("_cind", "_ass"),
    )

    #  Get rows where there is no AssessmentAuthorisationDate when a CINclosureDate is recorded.
    condition = df_merged[AssessmentAuthorisationDate].isna()
    df_merged = df_merged[condition].reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CINclosureDate and AssessmentAuthorisationDate in each child (differentiated by LAchildID)
    # So, a combination of LAchildID, CINclosureDate and AssessmentAuthorisationDate identifies and error instance.
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[CINdetailsID],
            df_merged[CINclosureDate],
        )
    )

    # The merges were done on copies of df_cind and df_ass so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cind_issues = (
        df_cind.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cind")
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_ass_issues = (
        df_ass.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=CINdetails, columns=[CINclosureDate], row_df=df_cind_issues
    )
    rule_context.push_type_2(
        table=Assessments, columns=[AssessmentAuthorisationDate], row_df=df_ass_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_cind = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID1",
                "CINclosureDate": "30/06/2000",  # Pass
            },
            {
                "LAchildID": "child2",
                "CINdetailsID": "CDID2",
                "CINclosureDate": "29/08/2002",  # Fails (no Assessment Authorisation Date entered)
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "CDID3",
                "CINclosureDate": "25/10/2002",  # Pass
            },
            {
                "LAchildID": "child4",
                "CINdetailsID": "CDID4",
                "CINclosureDate": "10/06/2000",  # Pass
            },
            {
                "LAchildID": "child5",
                "CINdetailsID": "CDID5",
                "CINclosureDate": "15/11/2001",  # Fails (no Assessment Authorisation Date entered)
            },
            {
                "LAchildID": "child6",
                "CINdetailsID": "CDID5",
                "CINclosureDate": "15/11/2001",  # Passes, no Assessment module
            },
        ]
    )

    sample_ass = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID1",
                "AssessmentAuthorisationDate": "26/05/2000",  # Pass
            },
            {
                "LAchildID": "child2",
                "CINdetailsID": "CDID2",
                "AssessmentAuthorisationDate": pd.NA,  # Fails (no date entered)
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "CDID3",
                "AssessmentAuthorisationDate": "26/05/2000",  # Pass
            },
            {
                "LAchildID": "child4",
                "CINdetailsID": "CDID4",
                "AssessmentAuthorisationDate": "30/05/2000",  # Pass
            },
            {
                "LAchildID": "child5",
                "CINdetailsID": "CDID5",
                "AssessmentAuthorisationDate": pd.NA,  # Fails (no date entered)
            },
        ]
    )
    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cind[CINclosureDate] = pd.to_datetime(
        sample_cind[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_ass[AssessmentAuthorisationDate] = pd.to_datetime(
        sample_ass[AssessmentAuthorisationDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            CINdetails: sample_cind,
            Assessments: sample_ass,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the Reviews columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace Reviews with the name of your table.
    issue_table = issues.table
    assert issue_table == Assessments

    # check that the right columns were returned. Replace CPPreviewDate  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df

    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    "CDID2",  # CindetailsID
                    # CIN Closure Date
                    pd.to_datetime("29/08/2002", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child5",  # ChildID
                    "CDID5",  # CindetailsID
                    # CIN Closure Date
                    pd.to_datetime("15/11/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)
    assert (issue_rows["ROW_ID"] == expected_df["ROW_ID"]).all()
    assert (issue_rows["ERROR_ID"] == expected_df["ERROR_ID"]).all()

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8867"
    assert (
        result.definition.message
        == "CIN episode is shown as closed, however Assessment is not shown as completed"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildProtectionPlans = CINTable.ChildProtectionPlans
InitialCategoryOfAbuse = ChildProtectionPlans.InitialCategoryOfAbuse


# define characteristics of rule
@rule_definition(
    code="8905",
    module=CINTable.ChildProtectionPlans,
    message="Initial Category of Abuse code missing or invalid (see Category of Abuse table in CIN Census code set)",
    affected_fields=[InitialCategoryOfAbuse],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildProtectionPlans]
    """
    Where a Child Protection Plan module is present, <InitialCategoryOfAbuse> (N00113) must be present and be a valid code
    """
    abuse_cats = ["NEG", "PHY", "SAB", "EMO", "MUL"]

    # Initial Category Code is not in list.
    df = df[
        (~df[InitialCategoryOfAbuse].isin(abuse_cats))
        | df[InitialCategoryOfAbuse].isna()
    ]

    failing_indices = df.index

    rule_context.push_issue(
        table=ChildProtectionPlans, field=InitialCategoryOfAbuse, row=failing_indices
    )


def test_validate():
    fake_cats = ["NEG", "AAA", "PHY", pd.NA, "BBB", "EMO", "MUL"]

    fake_dataframe = pd.DataFrame({InitialCategoryOfAbuse: fake_cats})

    result = run_rule(validate, {ChildProtectionPlans: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 3

    assert issues == [
        IssueLocator(CINTable.ChildProtectionPlans, InitialCategoryOfAbuse, 1),
        IssueLocator(CINTable.ChildProtectionPlans, InitialCategoryOfAbuse, 3),
        IssueLocator(CINTable.ChildProtectionPlans, InitialCategoryOfAbuse, 4),
    ]

    assert result.definition.code == "8905"
    assert (
        result.definition.message
        == "Initial Category of Abuse code missing or invalid (see Category of Abuse table in CIN Census code set)"
    )



"""
Rule number: '4004'
Module: CIN plan dates
Rule details: Within a <CINDetails> module, there must be only one <CINplanDates> group where the <CINPlanEnd Date> (N00690) is missing
Rule message: This child is showing more than one open CIN Plan, i.e. with no End Date

"""

from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.rules.cin2022_23.rule_8925 import LAchildID
from cin_validator.test_engine import run_rule

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID
CINPlanEndDate = CINplanDates.CINPlanEndDate
CINdetailsID = CINplanDates.CINdetailsID


@rule_definition(
    code="4004",
    module=CINTable.CINplanDates,
    message="This child is showing more than one open CIN Plan, i.e. with no End Date",
    affected_fields=[CINPlanEndDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINplanDates]
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    # There must be only one <CINplanDates> group where the <CINPlanEnd Date> (N00690) is missing
    df_check = df.copy()
    df_check = df_check[df_check[CINPlanEndDate].isna()]

    # Convert NAs to 1 and count by child
    df_check[CINPlanEndDate].fillna(1, inplace=True)
    df_check = (
        df_check.groupby([LAchildID, CINdetailsID])[CINPlanEndDate]
        .count()
        .reset_index()
    )

    # Find where there is more than 1 open end date
    df_check = df_check[df_check[CINPlanEndDate] > 1]
    issue_ids = tuple(zip(df_check[LAchildID], df_check[CINdetailsID]))

    df["ERROR_ID"] = tuple(zip(df[LAchildID], df[CINdetailsID]))
    df_issues = df[df.ERROR_ID.isin(issue_ids)]

    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    rule_context.push_type_3(
        table=CINplanDates, columns=[CINPlanEndDate], row_df=df_issues
    )


def test_validate():
    sample_CINplanDates = pd.DataFrame(
        [  # child1
            {
                LAchildID: "child1",
                CINdetailsID: "CINdetailsID1",
                CINPlanEndDate: pd.NA,  # 0 first nan date in group
            },
            {
                LAchildID: "child1",
                CINdetailsID: "CINdetailsID1",
                CINPlanEndDate: pd.NA,  # 1 second nan date in group
            },
            {  # won't be flagged because there is not more than one nan authorisation date in this group.
                LAchildID: "child1",
                CINdetailsID: "CINdetailsID2",
                CINPlanEndDate: pd.NA,  # 2
            },
            # child2
            {
                LAchildID: "child2",
                CINdetailsID: "CINdetailsID1",
                CINPlanEndDate: "26/05/2021",  # 3 ignored. not nan
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "CINdetailsID2",
                CINPlanEndDate: pd.NA,  # 4 first nan date in group
            },
            {  # fail
                LAchildID: "child2",
                CINdetailsID: "CINdetailsID2",
                CINPlanEndDate: pd.NA,  # 5 second nan date in group
            },
        ]
    )
    sample_CINplanDates[CINPlanEndDate] = pd.to_datetime(
        sample_CINplanDates[CINPlanEndDate],
        format="%d/%m/%Y",
        errors="coerce",
    )

    result = run_rule(validate, {CINplanDates: sample_CINplanDates})

    issues_list = result.type3_issues

    assert len(issues_list) == 1
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanEndDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "CINdetailsID1",
                ),
                "ROW_ID": [0, 1],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "CINdetailsID2",
                ),
                "ROW_ID": [4, 5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4004"
    assert (
        result.definition.message
        == "This child is showing more than one open CIN Plan, i.e. with no End Date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

Disabilities = CINTable.Disabilities
Disability = Disabilities.Disability
LAchildID = Disabilities.LAchildID


# define characteristics of rule
@rule_definition(
    code="2887Q",
    rule_type=RuleType.QUERY,
    module=CINTable.Disabilities,
    message="Please check and either amend or provide a reason: Less than 8 disability codes have been used in your return",
    affected_fields=[Disability],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Disabilities]

    # LOGIC
    # Excluding children where <Disability> (N00099) is ‘NONE’, the number of different disability codes used by a LA should be more than 7.
    # Validation should be triggered at LA level, not child level, if the LA has only used 7 or fewer different disability codes in their return.

    # remove "NONE" values
    condition = df[Disability] == "NONE"
    df = df[condition]

    # .nunique() will include NaN values so .count() is used instead which excludes NaNs.
    num_disabilities = df[Disability].count()
    if num_disabilities <= 7:
        rule_context.push_la_level(
            rule_context.definition.code, rule_context.definition.message
        )
    else:
        pass


def test_validate():
    sample_disabilities = pd.DataFrame(
        [
            {
                LAchildID: "child1",
                Disability: 0,
            },
            {LAchildID: "child1", Disability: "aaaa"},
            {LAchildID: "child2", Disability: "bbbb"},
            {LAchildID: "child3", Disability: "aaaa"},  # duplicate
            {LAchildID: "child4", Disability: "cc"},
            {LAchildID: "child5", Disability: "d"},
            {LAchildID: "child6", Disability: pd.NA},  # not counted
            {LAchildID: "child7", Disability: "f"},
            {LAchildID: "child8", Disability: "NONE"},  # ignored: Disability is NONE
            {LAchildID: "child9", Disability: "aaaa"},  # duplicate
        ]
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Disabilities: sample_disabilities})

    assert result.la_issues == (
        "2887Q",
        "Please check and either amend or provide a reason: Less than 8 disability codes have been used in your return",
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="1570",
    module=CINTable.ChildIdentifiers,
    message="LA Child ID must not be longer than 20 characters",
    affected_fields=[LAchildID],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers].copy()
    more_than_20_characters = df[df["LAchildID"].str.len() > 20]

    failing_indices = more_than_20_characters.index

    rule_context.push_issue(
        table=ChildIdentifiers, field=LAchildID, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    # Sidenote: a typical Header table will only have one row.
    sample_ci = pd.DataFrame(
        {
            "LAchildID": [
                "XXXXXXXXXXXXXXXXXXXX",  # 20
                "XXXXXXXXXXXXXXXXXXX",  # 19
                "XXXXXXXXXXXXXXXXXXXXX",  # 21 characters, fail
            ]
        }
    )

    result = run_rule(validate, {ChildIdentifiers: sample_ci})

    issues = list(result.issues)
    # Intended fail points in data.
    assert len(issues) == 1
    # Intended failures of test data by index.
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 2),
    ]

    # Checks rule code and message are correct.
    assert result.definition.code == "1570"
    assert (
        result.definition.message == "LA Child ID must not be longer than 20 characters"
    )



# this is similar to rule 8890

from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
LAchildID = Assessments.LAchildID
CINdetailsID = Assessments.CINdetailsID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8863Q",
    module=CINTable.Assessments,
    rule_type=RuleType.QUERY,
    message="An Assessment is shown as starting when there is another Assessment ongoing.",
    affected_fields=[AssessmentActualStartDate, AssessmentAuthorisationDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Name of the table required.
    df_ass = data_container[Assessments].copy()
    df_ass_2 = data_container[Assessments].copy()

    # Before you begin, rename the index and make it a column, so that the initial row positions can be kept intact.
    df_ass.index.name = "ROW_ID"
    df_ass_2.index.name = "ROW_ID"

    df_ass.reset_index(inplace=True)
    df_ass_2.reset_index(inplace=True)

    # ReferenceDate exists in the header table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Within one <CINdetails> group, where present each <AssessmentActualStartDate> (N00159) must not fall on or between:
    # a) the <AssessmentActualStartDate> (N00159) and <AssessmentAuthorisation Date (N00160) of any other <Assessments> group that has an <AssessmentAuthorisationDate> (N00160)
    # OR
    # b) the <AssessmentActualStartDate> (N00159) and the <ReferenceDate> (N00603) where the <AssessmentAuthorisationDate> (N00160) is missing

    df_ass = df_ass[df_ass[AssessmentActualStartDate].notna()]
    df_ass_2 = df_ass_2[df_ass_2[AssessmentActualStartDate].notna()]

    #  Merge tables to test for overlaps
    df_merged = df_ass.merge(
        df_ass_2,
        on=[LAchildID, CINdetailsID],
        how="left",
        suffixes=("_ass", "_ass2"),
    )

    # Prevent Assessments from being compared to themselves.
    same_start = (
        df_merged["AssessmentActualStartDate_ass"]
        == df_merged["AssessmentActualStartDate_ass2"]
    )
    same_end = (
        df_merged["AssessmentAuthorisationDate_ass"]
        == df_merged["AssessmentAuthorisationDate_ass2"]
    ) | (
        df_merged[
            "AssessmentAuthorisationDate_ass"
        ].isna()  # nans are checked separately because they are not considered equal by ==
        & df_merged["AssessmentAuthorisationDate_ass2"].isna()
    )
    duplicate = same_start & same_end
    df_merged = df_merged[~duplicate]

    # Determine whether assessment overlaps with another assessment
    ass_started_after_start = (
        df_merged["AssessmentActualStartDate_ass"]  # 1 starts later than 2 starts
        >= df_merged["AssessmentActualStartDate_ass2"]
    )
    ass_started_before_end = (
        df_merged["AssessmentActualStartDate_ass"]  # 1 starts earlier than 2 finishes
        <= df_merged["AssessmentAuthorisationDate_ass2"]
    ) & df_merged["AssessmentAuthorisationDate_ass2"].notna()
    ass_started_before_refdate = (
        df_merged["AssessmentActualStartDate_ass"] <= reference_date
    ) & df_merged["AssessmentAuthorisationDate_ass2"].isna()

    df_merged = df_merged[
        ass_started_after_start & (ass_started_before_end | ass_started_before_refdate)
    ].reset_index()

    # create an identifier for each error instance.
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[CINdetailsID],
            df_merged["AssessmentActualStartDate_ass"],
        )
    )

    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_ass_issues = (
        df_ass.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_ass_2_issues = (
        df_ass_2.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_ass2")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=Assessments, columns=[AssessmentActualStartDate], row_df=df_ass_issues
    )
    rule_context.push_type_3(
        table=Assessments,
        columns=[AssessmentActualStartDate, AssessmentAuthorisationDate],
        row_df=df_ass_2_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_ass = pd.DataFrame(
        [  # child1
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "26/05/2000",  # 0 Pass: not between "26/08/2000" and "31/03/2001"
                "AssessmentAuthorisationDate": "26/10/2000",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "26/08/2000",  # 1 Fail: between "26/05/2000" and "26/10/2000"
                "AssessmentAuthorisationDate": pd.NA,
            },
            {
                "LAchildID": "child2",  # 2 alone in cin group: not compared
                "CINdetailsID": "cinID2",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": "25/10/2000",
            },
            {
                "LAchildID": "child2",  # 3 alone in cin group: not compared
                "CINdetailsID": "cinID22",
                "AssessmentActualStartDate": "26/10/2000",
                "AssessmentAuthorisationDate": "26/12/2000",
            },
            # child3
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID3",
                "AssessmentActualStartDate": "26/05/2000",  # 4 Pass: not between "26/08/2000" and "26/10/2000"
                "AssessmentAuthorisationDate": "26/10/2001",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID3",
                "AssessmentActualStartDate": "26/08/2000",  # 5 Fail: between "26/05/2000" and "26/10/2001"
                "AssessmentAuthorisationDate": "26/10/2000",
            },
            # child4
            {
                "LAchildID": "child4",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "26/10/2000",  # 6 Fail: between "26/09/2000" and ReferenceDate
                "AssessmentAuthorisationDate": "31/03/2001",
            },
            {
                "LAchildID": "child4",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "26/09/2000",  # 7 Pass: not between "26/10/2000" and "31/03/2001"
                "AssessmentAuthorisationDate": pd.NA,
            },
            {
                "LAchildID": "child5",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "01/03/2000",
                "AssessmentAuthorisationDate": "01/04/2000",
            },
            {
                "LAchildID": "child5",
                "CINdetailsID": "cinID1",
                "AssessmentActualStartDate": "01/09/2000",
                "AssessmentAuthorisationDate": "01/10/2000",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_ass[AssessmentActualStartDate] = pd.to_datetime(
        sample_ass[AssessmentActualStartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_ass[AssessmentAuthorisationDate] = pd.to_datetime(
        sample_ass[AssessmentAuthorisationDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            Assessments: sample_ass,
            Header: sample_header,
        },
    )

    issues_list = result.type3_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the Reviews columns because that's the second thing pushed above.
    issues = issues_list[0]

    # get table name and check it. Replace Reviews with the name of your table.
    issue_table = issues.table
    assert issue_table == Assessments

    issue_columns = issues.columns
    assert issue_columns == [AssessmentActualStartDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3

    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID1",
                    pd.to_datetime("26/08/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID3",
                    pd.to_datetime("26/08/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "cinID1",
                    pd.to_datetime("26/10/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [6],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8863Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8863Q"
    assert (
        result.definition.message
        == "An Assessment is shown as starting when there is another Assessment ongoing."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
CINclosureDate = CINdetails.CINclosureDate
LAchildID = CINdetails.LAchildID


# define characteristics of rule
@rule_definition(
    code="8630",
    module=CINTable.CINdetails,
    message="CIN Closure Date is before CIN Referral Date for the same CIN episode",
    affected_fields=[CINreferralDate, CINclosureDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]

    # If present <CINclosureDate> (N00102) must be on or after <CINreferralDate> (N00100) in the same <CINdetails> instance

    df.index.name = "ROW_ID"

    df.query(
        "(CINclosureDate < CINreferralDate) and CINclosureDate.notna() and CINreferralDate.notna()",
        inplace=True,
    )

    df_issues = df.reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[CINreferralDate],
            df_issues[CINclosureDate],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=CINdetails,
        columns=[CINreferralDate, CINclosureDate],
        row_df=df_issues,
    )


def test_validate():
    IDS_are = [
        "AAAAAAAA",
        "BBBBBBBBB",
        "CCCCCCCCCCC",
        "DDDDDDDDD",
        "EEEE",
        "FFFFFFFFF",
        "GGGGGGGGGG",
        "HHHH",
    ]
    RefStart = [
        "01-01-2020",
        "01-02-2020",
        "01-03-2020",
        "15-01-2020",
        pd.NA,
        "01-07-2020",
        "15-01-2020",
        pd.NA,
    ]
    CINClose = [
        "01-01-2020",
        "01-01-2020",  #  Fails as Close before RefStart
        "01-03-2020",
        "17-01-2020",
        pd.NA,
        "01-01-2020",  #  Fails as Close before RefStart
        "15-01-2020",
        "01-01-2020",
    ]
    fake_dataframe = pd.DataFrame(
        {
            LAchildID: IDS_are,
            CINreferralDate: RefStart,
            CINclosureDate: CINClose,
        }
    )

    fake_dataframe[CINreferralDate] = pd.to_datetime(
        fake_dataframe[CINreferralDate], format=r"%d-%m-%Y", errors="coerce"
    )
    fake_dataframe[CINclosureDate] = pd.to_datetime(
        fake_dataframe[CINclosureDate], format=r"%d-%m-%Y", errors="coerce"
    )

    result = run_rule(validate, {CINdetails: fake_dataframe})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate, CINclosureDate]

    issue_rows = issues.row_df

    assert len(issue_rows) == 2

    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "BBBBBBBBB",
                    pd.to_datetime("01-02-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "FFFFFFFFF",
                    pd.to_datetime("01-07-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8630"
    assert (
        result.definition.message
        == "CIN Closure Date is before CIN Referral Date for the same CIN episode"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPendDate = ChildProtectionPlans.CPPendDate
CPPstartDate = ChildProtectionPlans.CPPstartDate
LAchildID = ChildProtectionPlans.LAchildID


# define characteristics of rule
@rule_definition(
    code="8925",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child Protection Plan End Date earlier than Start Date",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[CPPstartDate, CPPendDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[ChildProtectionPlans]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If present <CPPendDate> (N00115) must be on or after the <CPPstartDate> (N00105)
    condition = df[CPPendDate] < df[CPPstartDate]
    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error. In this case,
    # - If only LAchildID is used as an identifier, multiple instances of the error on a child will be understood as 1 instance.
    # We don't want that because in reality, a child can have multiple instances of an error.
    # - If we use the LAchildID-CPPstartDate combination, that artificially cancels out the instances where a start date repeats for the same child.
    # Another rule checks for that condition. Not this one.
    # - It is very unlikely that a combination of LAchildID-CPPstartDate-CPPendDate will repeat in the DataFrame.
    # Hence, it can be used as a unique identifier of the row.

    # Replace CPPstartDate and CPPendDate below with the columns concerned in your rule.
    link_id = tuple(
        zip(df_issues[LAchildID], df_issues[CPPstartDate], df_issues[CPPendDate])
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=ChildProtectionPlans, columns=[CPPstartDate, CPPendDate], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_protection_plans = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "26/05/2001",
            },
            {
                "LAchildID": "child3",
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "26/05/1999",
            },  # 2 error: end is before start
            {
                "LAchildID": "child3",
                "CPPstartDate": "26/05/2000",
                "CPPendDate": pd.NA,
            },
            {
                "LAchildID": "child4",
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "25/05/2000",
            },  # 4 error: end is before start
            {
                "LAchildID": "child5",
                "CPPstartDate": pd.NA,
                "CPPendDate": pd.NA,
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    child_protection_plans[CPPstartDate] = pd.to_datetime(
        child_protection_plans[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    child_protection_plans[CPPendDate] = pd.to_datetime(
        child_protection_plans[CPPendDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildProtectionPlans: child_protection_plans})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    # get table name and check it. Replace ChildProtectionPlans with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildProtectionPlans

    # check that the right columns were returned. Replace CPPstartDate and CPPendDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CPPstartDate, CPPendDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/1999", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8925' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8925"
    assert (
        result.definition.message
        == "Child Protection Plan End Date earlier than Start Date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINplanDates = CINTable.CINplanDates
Reviews = CINTable.Reviews
ChildProtectionPlans = CINTable.ChildProtectionPlans
CPPendDate = ChildProtectionPlans.CPPendDate
LAchildID = CINplanDates.LAchildID
CPPreviewDate = Reviews.CPPreviewDate
CPPID = Reviews.CPPID
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate
CPPendDate = ChildProtectionPlans.CPPendDate


@rule_definition(
    code="4003",
    module=CINTable.Reviews,
    message="A CPP review date is shown as being held at the same time as an open CIN Plan.",
    affected_fields=[
        CINPlanStartDate,
        CINPlanEndDate,
        CPPreviewDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpp = data_container[ChildProtectionPlans].copy()
    df_cin = data_container[CINplanDates].copy()
    df_reviews = data_container[Reviews].copy()

    df_reviews.index.name = "ROW_ID"
    df_cpp.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_reviews.reset_index(inplace=True)
    df_cpp.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # Within a <CINDetails> module, no <CPPReviewDate> (N00116) can fall between any
    # <CINPlanStartdate> (N00689) or <CINPlanEndDate> (N00690) unless <CPPReviewDate> is equal to <CPPendDate> (N00115)
    df_cpp = df_cpp.merge(
        df_reviews, on=["LAchildID", "CPPID"], how="left", suffixes=("", "_reviews")
    )

    df_merged = df_cin.merge(
        df_cpp,
        on=["LAchildID"],
        how="left",
        suffixes=("_cin", "_cpp"),
    )

    cin_start_after_cin_start = df_merged[CPPreviewDate] >= df_merged[CINPlanStartDate]
    cin_start_before_cin_end = (
        df_merged[CPPreviewDate] < df_merged[CINPlanEndDate]
    ) & df_merged[CPPendDate].notna()
    cp_review_is_end = (df_merged[CPPreviewDate] == df_merged[CPPendDate]) & df_merged[
        CPPendDate
    ].notna()

    df_merged = df_merged[
        (cin_start_after_cin_start & cin_start_before_cin_end) & (~cp_review_is_end)
    ].reset_index()

    df_merged["ERROR_ID"] = tuple(zip(df_merged[LAchildID], df_merged[CPPreviewDate]))
    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_reviews_issues = (
        df_reviews.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_reviews")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildProtectionPlans,
        columns=[CPPendDate],
        row_df=df_cpp_issues,
    )
    rule_context.push_type_2(
        table=CINplanDates, columns=[CINPlanStartDate], row_df=df_cin_issues
    )
    rule_context.push_type_2(
        table=Reviews, columns=[CPPreviewDate], row_df=df_reviews_issues
    )


def test_validate():
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPendDate": "30/05/2001",  # Fail
                "CPPID": "cinID1",
            },
            {
                "LAchildID": "child2",
                "CPPendDate": pd.NA,
                "CPPID": "cinID1",
            },
            {
                "LAchildID": "child2",
                "CPPendDate": "29/05/2001",  # ignore: CPPReviewDate == CPPendDate
                "CPPID": "cinID2",
            },
            {
                "LAchildID": "child2",
                "CPPendDate": pd.NA,
                "CPPID": "cinID4",
            },
            {
                "LAchildID": "child3",
                "CPPendDate": "30/10/2001",
                "CPPID": "cinID5",
            },
            {
                "LAchildID": "child5",
                "CPPendDate": pd.NA,
                "CPPID": "cinID6",
            },
        ]
    )
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINPlanStartDate": "04/04/2000",  # fail: 29/05/2001
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "28/05/2000",  # ignore: CPPReviewDate == CPPendDate 29/05/2001
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child3",
                "CINPlanStartDate": "30/05/2000",  # pass: "29/05/2004"
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child4",
                "CINPlanStartDate": "04/06/2000",  # pass: "29/05/2004"
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child5",
                "CINPlanStartDate": "30/06/2000",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/10/2000",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/02/2001",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/03/2001",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child3",
                "CINPlanStartDate": "30/10/2001",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child4",
                "CINPlanStartDate": "04/06/2000",
                "CINPlanEndDate": "01/06/2002",
            },
            {
                "LAchildID": "child5",
                "CINPlanStartDate": "31/03/2001",
                "CINPlanEndDate": "01/06/2002",
            },
        ]
    )
    sample_reviews = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CPPID": "cinID1",
                "CPPreviewDate": "29/05/2001",
            },  # fail
            {
                "LAchildID": "child2",
                "CPPID": "cinID2",
                "CPPreviewDate": "29/05/2001",
            },  # ignore
            {
                "LAchildID": "child3",
                "CPPID": "cinID3",
                "CPPreviewDate": "29/05/2004",
            },  # pass
            {
                "LAchildID": "child4",
                "CPPID": "cinID4",
                "CPPreviewDate": "29/05/2004",
            },  # pass
            {
                "LAchildID": "child5",
                "CPPID": "cinID5",
                "CPPreviewDate": "29/05/2004",
            },  # pass
            {
                "LAchildID": "child6",
                "CPPID": "cinID6",
                "CPPreviewDate": "29/05/2004",
            },  # ignore: not present in cin table
        ]
    )

    sample_reviews[CPPreviewDate] = pd.to_datetime(
        sample_reviews[CPPreviewDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cpp[CPPendDate] = pd.to_datetime(
        sample_cpp[CPPendDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINPlanStartDate] = pd.to_datetime(
        sample_cin[CINPlanStartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINPlanEndDate] = pd.to_datetime(
        sample_cin[CINPlanEndDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            CINplanDates: sample_cin,
            Reviews: sample_reviews,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 3
    issues = issues_list[2]

    issue_table = issues.table
    assert issue_table == Reviews

    issue_columns = issues.columns
    assert issue_columns == [CPPreviewDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("29/05/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            }
        ]
    )

    assert issue_rows.equals(expected_df)
    assert result.definition.code == "4003"
    assert (
        result.definition.message
        == "A CPP review date is shown as being held at the same time as an open CIN Plan."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildIdentifiers = CINTable.ChildIdentifiers
CINdetails = CINTable.CINdetails
LAchildID = ChildIdentifiers.LAchildID


PersonBirthDate = ChildIdentifiers.PersonBirthDate
CINclosureDate = CINdetails.CINclosureDate

# Reference date in header is needed to define the period of census.
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 8775Q
    code="8775Q",
    rule_type=RuleType.QUERY,
    # replace ChildIdentifiers with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.ChildIdentifiers,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Please check and either amend data or provide a reason: Child is over 25 years old",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        PersonBirthDate,
        CINclosureDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_ci = data_container[ChildIdentifiers].copy()
    df_cin = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_ci.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_ci.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # get collection period
    df_ref = data_container[Header]
    ref_date = df_ref[ReferenceDate].iloc[0]

    # lOGIC
    # <PersonBirthDate> (N00066) is before (<ReferenceDate> (N00603) minus 25 years) AND
    # {<CINClosureDate> (N00102) is after (<PersonBirthDate> plus 25 years) OR <CINClosureDate> is NULL}
    over_25 = df_ci[PersonBirthDate] < (ref_date - pd.DateOffset(years=25))
    df_ci = df_ci[over_25]

    merged_df = df_ci.merge(
        df_cin,
        on=[
            LAchildID,
        ],
        suffixes=["_ci", "_cin"],
    )

    condition = (
        (
            merged_df["CINclosureDate"]
            > (merged_df[PersonBirthDate] + pd.DateOffset(years=25))
        )
        | (merged_df["CINclosureDate"].isna())
        | (merged_df["CINclosureDate"] == "NULL")
    )

    merged_df = merged_df[condition].reset_index()

    # create an identifier for each error instance.
    merged_df["ERROR_ID"] = tuple(zip(merged_df[LAchildID], merged_df[PersonBirthDate]))

    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_ci_issues = (
        df_ci.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_ci")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonBirthDate], row_df=df_ci_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[CINclosureDate], row_df=df_cin_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_ci = pd.DataFrame(
        [
            {"LAchildID": "child1", "PersonBirthDate": "01/01/1880"},
            {"LAchildID": "child2", "PersonBirthDate": "01/01/1880"},
            {
                "LAchildID": "child3",
                "PersonBirthDate": "01/01/2000",
            },  # ignore: not 25 years before refdate
            {
                "LAchildID": "child4",
                "PersonBirthDate": "01/01/1800",
            },  # ignore: CINclosureDate not up to 25years after birthdate
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {  # 0 fail
                "LAchildID": "child1",
                "CINclosureDate": "01/01/2000",
            },
            {  # 1 fail
                "LAchildID": "child2",
                "CINclosureDate": "NULL",
            },
            {  # 2 pass
                "LAchildID": "child2",
                "CINclosureDate": "26/10/1890",
            },
            {  # 3 pass
                "LAchildID": "child3",
                "CINclosureDate": "26/10/1999",
            },
            {  # 4 pass
                "LAchildID": "child4",
                "CINclosureDate": "26/10/1804",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_ci["PersonBirthDate"] = pd.to_datetime(
        sample_ci["PersonBirthDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin_details["CINclosureDate"] = pd.to_datetime(
        sample_cin_details["CINclosureDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_header = pd.DataFrame(
        [
            {
                ReferenceDate: pd.to_datetime(
                    "31/03/2001", format="%d/%m/%Y", errors="coerce"
                )
            }
        ]  # the census start date here will be 01/04/2000
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_ci,
            CINdetails: sample_cin_details,
            Header: sample_header,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    issues = issues_list[1]

    # get table name and check it. Replace Section47 with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINclosureDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    pd.to_datetime("01/01/1880", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("01/01/1880", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8775Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8775Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Child is over 25 years old"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

Section47 = CINTable.Section47
DateOfInitialCPC = Section47.DateOfInitialCPC

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID


@rule_definition(
    code="1104",
    module=CINTable.Section47,
    message="The date of the initial child protection conference cannot be before the referral date",
    affected_fields=[
        CINreferralDate,
        DateOfInitialCPC,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cin = data_container[CINdetails]
    df_47 = data_container[Section47]

    df_cin.index.name = "ROW_ID"
    df_47.index.name = "ROW_ID"

    df_cin.reset_index(inplace=True)
    df_47.reset_index(inplace=True)

    # Where present, the <DateOfInitialCPC> (N00110) should be on or after <CINreferralDate> (N00100)
    df_47 = df_47[df_47[DateOfInitialCPC].notna()]
    # get only relevant rows in df_47 (line above) and relevant columns in CIN
    # (line below: prevent the other DateOfInitialCPC from coming along in the merge else DateOfInitialCPC column name
    # will depend on whether the same name in present in the CINdetails table and that is out of scope for this rule.)
    df_cin_filtered = df_cin[["ROW_ID", LAchildID, CINdetailsID, CINreferralDate]]

    merged_df = df_47.copy().merge(
        df_cin_filtered,
        on=[LAchildID, CINdetailsID],
        how="left",
        suffixes=["_47", "_cin"],
    )

    # check that the the dates being compared existed in the same CIN event period and belong to the same child.
    condition = merged_df[DateOfInitialCPC] < merged_df[CINreferralDate]

    merged_df = merged_df[condition].reset_index()

    # create an identifier for each error instance.
    merged_df["ERROR_ID"] = tuple(
        zip(merged_df[LAchildID], merged_df[CINdetailsID], merged_df[DateOfInitialCPC])
    )

    # The merges were done on copies of df_47 and df_cin so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_47_issues = (
        df_47.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=Section47, columns=[DateOfInitialCPC], row_df=df_47_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_cin_issues
    )


def test_validate():
    sample_section47 = pd.DataFrame(
        [
            {  # 0 fail: before
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": "19/05/2000",
            },
            {  # 1 fail: before
                "LAchildID": "child1",
                "CINdetailsID": "cinID2",
                "DateOfInitialCPC": "26/05/2000",
            },
            {  # 2 ignore
                "LAchildID": "child2",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": pd.NA,
            },
            {  # 3 pass: after
                "LAchildID": "child3",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": "31/05/2003",
            },
            {  # 4 ignore: CINreferralDate is null
                "LAchildID": "child3",
                "CINdetailsID": "cinID2",
                "DateOfInitialCPC": "26/05/2000",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {  # 0 fail
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",
                "CINreferralDate": "26/10/2001",
            },
            {  # 1 fail
                "LAchildID": "child1",
                "CINdetailsID": "cinID2",
                "CINreferralDate": "13/06/2002",
            },
            {  # 2 ignore
                "LAchildID": "child2",
                "CINdetailsID": "cinID1",
                "CINreferralDate": "26/05/2000",
            },
            {  # 3 pass
                "LAchildID": "child3",
                "CINdetailsID": "cinID1",
                "CINreferralDate": "28/05/2000",
            },
            {  # 4 ignore
                "LAchildID": "child3",
                "CINdetailsID": "cinID2",
                "CINreferralDate": pd.NA,
            },
            {  # 5 ignore: doesn't match any ID when merged
                "LAchildID": "child3",
                "CINdetailsID": "cinID4",
                "CINreferralDate": "28/05/2000",
            },
        ]
    )
    sample_section47["DateOfInitialCPC"] = pd.to_datetime(
        sample_section47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin_details["CINreferralDate"] = pd.to_datetime(
        sample_cin_details["CINreferralDate"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            Section47: sample_section47,
            CINdetails: sample_cin_details,
        },
    )

    # Type 2 rule.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID
                    # corresponding DateofInitialCPC
                    pd.to_datetime("19/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child1",
                    "cinID2",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "1104"
    assert (
        result.definition.message
        == "The date of the initial child protection conference cannot be before the referral date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildIdentifiers = CINTable.ChildIdentifiers
CINdetails = CINTable.CINdetails

LAchildID = ChildIdentifiers.LAchildID
PersonBirthDate = ChildIdentifiers.PersonBirthDate
UPN = ChildIdentifiers.UPN
UPNunknown = ChildIdentifiers.UPNunknown
ReferralNFA = CINdetails.ReferralNFA

# Reference date in header is needed to define the period of census.
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 8770Q
    code="8770Q",
    rule_type=RuleType.QUERY,
    # replace ChildIdentifiers with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.ChildIdentifiers,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Please check and either amend data or provide a reason: UPN or reason UPN missing expected for a child who is more than 5 years old",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        UPNunknown,
        UPNunknown,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_cid = data_container[ChildIdentifiers].copy()
    df_cin = data_container[CINdetails].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cid.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cid.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # get collection period
    header = data_container[Header]
    ref_date = header[ReferenceDate].iloc[0]
    ref_date_minus6 = ref_date - pd.DateOffset(years=6)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <ReferralNFA> (N00112) = 0 or false
    # and <PersonBirthDate> (N00066) is earlier than [<ReferenceDate> (N00603) minus 6 years]
    # then either <UPN> (N00001) or a valid <UPNunknown> (N00135) should be present

    # get only CINdetails where ReferralNFA is 0 or false
    df_cin = df_cin[
        (df_cin[ReferralNFA] == "0") | (df_cin[ReferralNFA].str.lower() == "false")
    ]

    # get only ChildIdentifiers where child is old enough to need a UPN
    df_cid = df_cid[(df_cid[PersonBirthDate] <= ref_date_minus6)]

    # merge ChildIdentifiers with filtered CINdetails and take only those that match
    df_merged = df_cid.merge(
        df_cin[[LAchildID, ReferralNFA]],
        on=[LAchildID],
        how="left",
        suffixes=["", "_cin"],
        indicator=True,
    )

    df_merged = df_merged[df_merged["_merge"] == "both"]

    # get only rows where UPN is not provided and UPNunknown is not a valid code
    no_upn = df_merged[UPN].isna()

    upnunknown_reasons = [
        # "UN1", - excluding UN1 as "Child is not of school age" is not applicable here
        "UN2",
        "UN3",
        "UN4",
        "UN5",
        "UN6",
        "UN7",
    ]

    valid_upnunknown = df_merged[UPNunknown].str.upper().isin(upnunknown_reasons)

    # get all the data that fits the failing condition.
    df_merged = df_merged[no_upn & ~valid_upnunknown].reset_index()

    # create an identifier for each error instance.
    df_merged["ERROR_ID"] = tuple(zip(df_merged[LAchildID]))

    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cid_issues = (
        df_cid.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[UPN, UPNunknown], row_df=df_cid_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_cid = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Ignore - under 5 years old
                "PersonBirthDate": "26/05/2000",
                "UPN": pd.NA,
                "UPNunknown": pd.NA,
            },
            {
                "LAchildID": "child2",  # 1 Pass - over 5 with UPN
                "PersonBirthDate": "26/05/1990",
                "UPN": "AAA",
                "UPNunknown": pd.NA,
            },
            {
                "LAchildID": "child3",  # 2 Pass - over 5 with valid UPNunknown
                "PersonBirthDate": "26/05/1990",
                "UPN": pd.NA,
                "UPNunknown": "UN2",
            },
            {
                "LAchildID": "child4",  # 3 Pass - over 5 with valid UPNunknown (lower case)
                "PersonBirthDate": "26/05/1990",
                "UPN": pd.NA,
                "UPNunknown": "un5",
            },
            {
                "LAchildID": "child5",  # 4 Fail - over 5 and no UPN or valid UPNunknown
                "PersonBirthDate": "26/05/1990",
                "UPN": pd.NA,
                "UPNunknown": pd.NA,
            },
            {
                "LAchildID": "child6",  # 5 Fail - over 5 and no UPN or valid UPNunknown (UN1 is not valid for >5yrs)
                "PersonBirthDate": "26/05/1990",
                "UPN": pd.NA,
                "UPNunknown": "UN1",
            },
            {
                "LAchildID": "child7",  # 6 Ignore - over 5 and no UPN or valid UPNunknown (UN8 is not valid) so should fail, BUT ReferralNFA doesn't meet requirement
                "PersonBirthDate": "26/05/1990",
                "UPN": pd.NA,
                "UPNunknown": "UN8",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {"LAchildID": "child1", "ReferralNFA": "0"},
            {"LAchildID": "child2", "ReferralNFA": "false"},
            {"LAchildID": "child3", "ReferralNFA": "FALSE"},
            {"LAchildID": "child4", "ReferralNFA": "0"},
            {"LAchildID": "child5", "ReferralNFA": "0"},
            {"LAchildID": "child6", "ReferralNFA": "0"},
            {"LAchildID": "child6", "ReferralNFA": "1"},
            {"LAchildID": "child7", "ReferralNFA": "1"},
        ]
    )
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cid[PersonBirthDate] = pd.to_datetime(
        sample_cid[PersonBirthDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_cid,
            CINdetails: sample_cin_details,
            Header: sample_header,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 1
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values.
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [UPN, UPNunknown]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": ("child5",),  # ChildID
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": ("child6",),  # ChildID
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8770Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8770Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: UPN or reason UPN missing expected for a child who is more than 5 years old"
    )



"""
Rule number: '8510'
Module: Child idenitifiers
Rule details: Each <LAchildID> (N00097) must be unique across all children within the same LA return. 

Note: This rule should be evaluated at LA-level for imported data

Rule message: More than one child record with the same LA Child ID

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="8510",
    module=CINTable.ChildIdentifiers,
    message="More than one child record with the same LA Child ID",
    affected_fields=[LAchildID],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]

    # Each <LAchildID> (N00097) must be unique across all children within the same LA return
    failing_indices = df[df.duplicated(subset=[LAchildID], keep=False)].index

    rule_context.push_issue(
        table=ChildIdentifiers, field=LAchildID, row=failing_indices
    )


def test_validate():
    child_identifiers = pd.DataFrame([[1234], [1234], [346546]], columns=[LAchildID])

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 0),
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 1),
    ]

    assert result.definition.code == "8510"
    assert (
        result.definition.message
        == "More than one child record with the same LA Child ID"
    )
